﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [Item("Linear", "Linear transformation x' = Scale * x + Offset")]
  [StorableClass]
  public class LinearTransformation : Transformation<double> {

    #region Parameters
    private FixedValueParameter<DoubleValue> ScaleParameter {
      get { return (FixedValueParameter<DoubleValue>)Parameters["Scale"]; }
    }

    private FixedValueParameter<DoubleValue> OffsetParameter {
      get { return (FixedValueParameter<DoubleValue>)Parameters["Offset"]; }
    }
    #endregion

    #region Properties
    public double Scale {
      get { return ScaleParameter.Value.Value; }
      set { ScaleParameter.Value.Value = value; }
    }

    public double Offset {
      get { return OffsetParameter.Value.Value; }
      set { OffsetParameter.Value.Value = value; }
    }
    #endregion

    #region Constructor, Cloning & Persistence
    public LinearTransformation()
      : base() {
      Parameters.Add(new FixedValueParameter<DoubleValue>("Scale", "Multiplicative factor (Slope)", new DoubleValue(1.0)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("Offset", "Additive factor (Intercept)", new DoubleValue(0.0)));
    }

    protected LinearTransformation(LinearTransformation original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new LinearTransformation(this, cloner);
    }

    [StorableConstructor]
    protected LinearTransformation(bool deserializing)
      : base(deserializing) {
    }
    #endregion

    public override IEnumerable<double> Apply(IEnumerable<double> data) {
      return Apply(data, Scale, Offset);
    }

    public override IEnumerable<double> InverseApply(IEnumerable<double> data) {
      return InverseApply(data, Scale, Offset);
    }


    public static IEnumerable<double> Apply(IEnumerable<double> data, double scale = 1.0, double offset = 0.0) {
      if (scale.IsAlmost(0.0))
        throw new InvalidOperationException($"Cannot transform with a {nameof(scale)} of zero because inverse transformation would be invalid.");

      return data.Select(x => scale * x + offset);
    }

    public static IEnumerable<double> InverseApply(IEnumerable<double> data, double scale = 1.0, double offset = 0.0) {
      if (scale.IsAlmost(0.0))
        throw new InvalidOperationException($"Cannot inverse transform with a {nameof(scale)} of zero.");

      return data.Select(x => (x - offset) / scale);
    }
  }
}